/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include "config.h"

#include <stdio.h>
#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <libintl.h>

#include "Semaphore.h"
#include "PThreads.h"
#include "sambascanner.h"

extern struct Options_Structure Options;
static struct Semaphore_Data ThreadCount_Semaphore;
static pthread_mutex_t Exec_Lock;
static pthread_mutex_t Memory_Lock;
static pthread_mutex_t Signal_Lock;
static pthread_mutex_t Processes_Lock;
static pthread_mutex_t Child_Lock;
static pthread_cond_t Child_Cond;

void Init_Semaphores() {
	Semaphore_Init(&ThreadCount_Semaphore, Options.ThreadCount, "ThreadCount_Semaphore");
	Safe_pthread_mutex_init(&Memory_Lock, "Memory_Lock");
	Safe_pthread_mutex_init(&Signal_Lock, "Signal_Lock");
	Safe_pthread_mutex_init(&Processes_Lock, "Processes_Lock");
	Safe_pthread_mutex_init(&Child_Lock, "Child_Lock");
	Safe_pthread_mutex_init(&Exec_Lock, "Exec_Lock");
	Safe_pthread_cond_init(&Child_Cond, "Child_Cond");
}

/*******************************************************************************************/
/* Lock and unlock the "exec" code right after forking                                     */
void Lock_Exec() {
	Safe_pthread_mutex_lock(&Exec_Lock);
}

void Unlock_Exec() {
	Safe_pthread_mutex_unlock(&Exec_Lock);
}
/*******************************************************************************************/


/*******************************************************************************************/
/* Thread is created/has exited                                                            */
void ThreadCnt_Inc() {
	Semaphore_P(&ThreadCount_Semaphore, 1);
}

void ThreadCnt_Dec() {
	Semaphore_V(&ThreadCount_Semaphore, 1);
}

void ThreadCnt_WaitAllFinished() {
	Semaphore_P(&ThreadCount_Semaphore, Options.ThreadCount);
	Semaphore_SetValue(&ThreadCount_Semaphore, Options.ThreadCount);
}
/*******************************************************************************************/


/*******************************************************************************************/
/* Lock/unlock the table of active processes                                               */
void Lock_ProcessTable() {
	Safe_pthread_mutex_lock(&Processes_Lock);
}

void Unlock_ProcessTable() {
	Safe_pthread_mutex_unlock(&Processes_Lock);
}
/*******************************************************************************************/

/*******************************************************************************************/
/* - Wait until there are children dead (waiting for process to finish)                    */
/* - Signal that there is a dead child (child handler does this)                           */
/* - Signal that we no longer wait for a process to finish when signalling was successful  */
/* - Check if the "running" flag is set, acquire lock so we cannot deadlock                */
void Sleep_Until_Children_Dead() {
	pthread_cond_wait(&Child_Cond, &Child_Lock); 
}

void Signal_Died_Children() {
	pthread_cond_broadcast(&Child_Cond);
}

void Accept_Dead_Child() {
	Safe_pthread_mutex_unlock(&Child_Lock);
}

void Child_Check_Condition() {
	Safe_pthread_mutex_lock(&Child_Lock);
}
/*******************************************************************************************/

void Free_Semaphores() {
	Safe_pthread_cond_destroy(&Child_Cond, "Child_Cond");
	Safe_pthread_mutex_destroy(&Exec_Lock, "Exec_Lock");
	Safe_pthread_mutex_destroy(&Child_Lock, "Child_Lock");
	Safe_pthread_mutex_destroy(&Processes_Lock, "Processes_Lock");
	Safe_pthread_mutex_destroy(&Signal_Lock, "Signal_Lock");
	Safe_pthread_mutex_destroy(&Memory_Lock, "Memory_Lock");
	Semaphore_Free(&ThreadCount_Semaphore, "ThreadCount_Semaphore");
}
