/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include "config.h"

#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <libsmbclient.h>

#include "PThreads.h"
#include "smblister.h"
#include "String.h"

#define DEBUG_LEVEL 	0
#define BUFSIZE			1024

static char *Default_Usernames[] = {
	"admin",
	"administrator",
	"guest",
	"gast",
	"anonymous",
	NULL
};

static char *Default_Passwords[] = {
	"",
	"admin",
	"administrator",
	"gast",
	"guest",
	"password",
	"passwort",
	"root",
	NULL
};

static int User_Index = 0;
static int PW_Index = 0;
static char Try_Default_Passwords = 0;

static char **Supplied_Usernames = NULL;
static char **Supplied_Passwords = NULL;
static int Number_Supplied_Usernames = 0;
static int Supplied_Index = 0;

static char *Username = NULL;
static char *Password = NULL;

static int Get_Next_Password() {
	static int Calls = 0;
	Calls++;

	if (Supplied_Index < Number_Supplied_Usernames) {
		Username = Supplied_Usernames[Supplied_Index];
		Password = Supplied_Passwords[Supplied_Index];
		Supplied_Index++;
		return 1;
	}
			
	if (Try_Default_Passwords == 1) {
		Username = Default_Usernames[User_Index];
		Password = Default_Passwords[PW_Index];
		PW_Index++;
		if (Default_Passwords[PW_Index] == NULL) {
			PW_Index = 0;
			User_Index++;
			if (Default_Usernames[User_Index] == NULL) {
				Try_Default_Passwords = 0;
			}
		}
		return 1;
	}

	if (Calls == 1) {
		/* So at least one combination gets returned */
		Username = "admin";
		Password = "";
		return 1;
	}

	return 0;
}

static void Authentication_Function(const char *pServer, const char *pShare, char *pWorkgroup, int workgroup_len, char *pUsername, int username_len, char *pPassword, int password_len) {
    strncpy(pUsername, Username, username_len);
    strncpy(pPassword, Password, password_len);
}

static int Process_Parameters(int argc, char **argv, char **Url) {
	int i;
	*Url = argv[argc - 1];
	for (i = 1; i < argc - 1; i++) {
		if (!strcmp(argv[i], "-g")) {
			Try_Default_Passwords = 1;
		} else if (!strcmp(argv[i], "-u")) {
			if (i == argc - 2) {
				fprintf(stderr, "Ran out of parameters for username/password.\n");
				return RETURN_PARAMETERS_USERPASS;
			}
			{
				char *User, *Pass;
				char *SavePtr;
				User = strtok_r(argv[i + 1], ",", &SavePtr);
				if (User == NULL) {
					fprintf(stderr, "Invalid form of username/password.\n");
					return RETURN_INVALID_USERPASS1;
				}
				Pass = strtok_r(NULL, ",", &SavePtr);
				if (Pass == NULL) {
					fprintf(stderr, "Invalid form of username/password.\n");
					return RETURN_INVALID_USERPASS2;
				}
				Number_Supplied_Usernames++;
				Supplied_Usernames = (char**)realloc(Supplied_Usernames, sizeof(char*) * Number_Supplied_Usernames);
				Supplied_Passwords = (char**)realloc(Supplied_Passwords, sizeof(char*) * Number_Supplied_Usernames);
				Supplied_Usernames[Number_Supplied_Usernames - 1] = User;
				Supplied_Passwords[Number_Supplied_Usernames - 1] = Pass;
			}
			i++;
		} else {
			fprintf(stderr, "Unknown parameter: %s\n", argv[i]);
			return RETURN_UNKNOWN_PARAMETER;
		}
	}
	return 0;
}

int main(int argc, char **argv) {
	int dir;
	struct smbc_dirent *Directory_Entry;
	char *BaseUrl;
	char Buffer[BUFSIZE];
	int i;
	StringList Dirlist;
	int Successful = 0;

	if ((argc < 2) || (!strcmp(argv[1], "--help"))) {
		fprintf(stderr, "Syntax: %s [Options] [Sharename]\n", argv[0]);
		fprintf(stderr, "\n");
		fprintf(stderr, "Known options:\n");
		fprintf(stderr, "    -u Usr,Pwd  Try this specific combination of username/password\n");
		fprintf(stderr, "    -g          Tries to guess common username/password combinations\n");
		fprintf(stderr, "    --help      Display this help page\n");
		fprintf(stderr, "\n");
		fprintf(stderr, "Example: %s smb://NEO/PICTURES\n", argv[0]);
		fprintf(stderr, "\n");
		fprintf(stderr, "Using libsmbclient version %s.\n", smbc_version());
		exit(RETURN_SYNTAX);
	}

	i = Process_Parameters(argc, argv, &BaseUrl);
	if (i) exit(i);

	if (smbc_init(Authentication_Function, DEBUG_LEVEL) != 0) {
		fprintf(stderr, "smblister: Could not initialize smbc-library.\n");
		exit(RETURN_SMBC_FAILURE);
	}

	/* Check authentication first */
	while (Get_Next_Password()) {
		dir = smbc_opendir(BaseUrl);
		if (dir < 0) {
			fprintf(stderr, "smblister: Opening base directory %s as %s/%s failed (%s)\n", BaseUrl, Username, Password, strerror(errno));
		} else {
			Successful = 1;
			smbc_closedir(dir);
			break;
		}
	}
	if (!Successful) {
		fprintf(stderr, "smblister: Giving up, ran out of passwords.\n");
		exit(RETURN_PASSWORDPROTECTED);
	}

	Init_StringList(&Dirlist);
	Add_String(&Dirlist, BaseUrl);
	for (i = 0; i < Dirlist.Entries; i++) {
		char *Url;
		Url = Dirlist.Entry[i];
		if ((dir = smbc_opendir(Url)) < 0) {
			fprintf(stderr, "smblister: Could not open %s (%s)\n", Url, strerror(errno));
			continue;
		}

		while ((Directory_Entry = smbc_readdir(dir)) != NULL) {
			if (Directory_Entry->smbc_type == SMBC_DIR) {
				if (strcmp(Directory_Entry->name, ".") && strcmp(Directory_Entry->name, "..")) {
					snprintf(Buffer, sizeof(Buffer), "%s/%s", Url, Directory_Entry->name);
					Add_String_Unique(&Dirlist, Buffer);
				}
			} else 	if (Directory_Entry->smbc_type == SMBC_FILE) {
				off_t Filesize;
				struct stat Filestats;
				snprintf(Buffer, sizeof(Buffer), "%s/%s", Url, Directory_Entry->name);
				if (!smbc_stat(Buffer, &Filestats)) {
					Filesize = Filestats.st_size;
				} else {
					Filesize = 0;
				}

				Date_To_String(Buffer, sizeof(Buffer), Filestats.st_mtime);

#if (_FILE_OFFSET_BITS == 64)
				printf("%s/%s|%s|%s|%lld|%s\n", Url, Directory_Entry->name, Username, Password, Filesize, Buffer);
#else
				printf("%s/%s|%s|%s|%ld|%s\n", Url, Directory_Entry->name, Username, Password, Filesize, Buffer);
#endif
			}
		}
		smbc_closedir(dir);
	}
	Destroy_StringList(&Dirlist);

	return RETURN_SUCCESS;
}

